<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Webcam Streaming</title>
</head>
<body>
    <h1>Webcam Streaming App</h1>
    <form method="post" action="/submit">
        <input type="text" name="username" placeholder="Enter username" required>
        <button type="submit" name="action" value="Start Streaming">Start Streaming</button>
        <button type="submit" name="action" value="Watch Stream">Watch Stream</button>
    </form>
</body>
</html>

